"""The SomaConnect class."""
from __future__ import annotations

import logging
from typing import Any

import aiohttp
import backoff

_LOGGER = logging.getLogger(__name__)


def clamp(num: int, min_value: int, max_value: int) -> int:
    """Clamp num to be between min_value and max_value."""
    return max(min(num, max_value), min_value)


class SomaConnect:
    """Represents a connection to SOMA Connect."""

    def __init__(self, host: str, port: int) -> None:
        """Initialise the connection."""
        self._host = host.removeprefix("http://")
        self._port = port
        self._url = f"http://{host}:{port}"
        self._version: str = ""
        self._devices: set[tuple[str, str, str, str]] = set()

    @property
    def devices(self) -> list[tuple[str, str, str, str]] | None:
        """Return a list of discovered shades."""
        if len(self._devices) > 0:
            return [device for device in self._devices]
        return None

    @property
    def version(self) -> str:
        """Return the SOMA Connect version."""
        return self._version

    @backoff.on_exception(
        backoff.expo, aiohttp.ClientError, max_tries=3, logger=_LOGGER
    )
    async def _get(self, endpoint: str, **kwargs: dict[str, Any]) -> dict[str, Any]:
        """Issue a request and return the JSON."""
        url = f"{self._url}/{endpoint}"
        params = kwargs if len(kwargs) > 0 else {}

        async with aiohttp.ClientSession(raise_for_status=False) as session:
            async with session.get(url, params=params) as response:
                json: dict[str, str] = await response.json()
                result: str = json.get("result", "")

                if result == "error":
                    json.pop("version", None)
                    json.pop("mac", None)

                if (version := json.get("version", None)) is not None:
                    self._version = version

                return json

    async def list_devices(self) -> list[tuple[str, str, str, str]] | None:
        """Return list of devices."""
        result = await self._get("list_devices")
        if (devices := result.get("shades", None)) is not None and len(devices) > 0:
            for device in devices:
                self._devices.add(
                    (device["name"], device["mac"], device["type"], device["gen"])
                )

            return self.devices

        return None

    async def open_shade(self, mac: str) -> bool:
        """Open the shade."""
        if mac is not None:
            result = await self._get(f"open_shade/{mac}")
            if result.get("result", None) == "success":
                return True
        return False

    async def close_shade(self, mac: str) -> bool:
        """Close the shade."""
        if mac is not None:
            result = await self._get(f"close_shade/{mac}")
            if result.get("result", None) == "success":
                return True
        return False

    async def stop_shade(self, mac: str) -> bool:
        """Stop the shade."""
        if mac is not None:
            result = await self._get(f"stop_shade/{mac}")
            if result.get("result", None) == "success":
                return True
        return False

    async def get_shade_position(self, mac: str) -> int | bool:
        """Get shade state."""
        if mac is not None:
            result = await self._get(f"get_shade_state/{mac}")
            if result.get("result", None) == "success":
                return result.get("position", False)
        return False

    async def set_shade_position(
        self,
        mac: str,
        position: int,
        close_upwards: bool | None = False,
        morning_mode: bool | None = False,
    ) -> bool:
        """Set shade position."""
        kwargs = {}
        if mac is not None and isinstance(position, int):
            position = clamp(position, 0, 100)

            if close_upwards is True:
                kwargs["close_upwards"] = 1
            if morning_mode is True:
                kwargs["morning_mode"] = 1
            result = await self._get(
                f"set_shade_position/{mac}/{str(position)}", **kwargs
            )
            if result.get("result", None) == "success":
                return True
        return False

    async def get_battery_level(self, mac: str) -> tuple[int, int] | bool:
        """Get battery level."""
        if mac is not None:
            result = await self._get(f"get_battery_level/{mac}")
            if result.get("result", None) == "success":
                battery_level = result.get("battery_level", False)
                battery_percentage = result.get("battery_percentage", False)
                return (battery_level, battery_percentage)
        return False

    async def get_light_level(self, mac: str) -> int | None:
        """Get light level."""
        if mac is not None:
            result = await self._get(f"get_light_level/{mac}")
            if result.get("result", None) == "success":
                light_level = result.get("light_level", False)
                return light_level
        return None



"""The Shade class."""
from __future__ import annotations

import datetime

from .soma_connect import SomaConnect


class SomaShade:
    """Represents a SOMA shade."""

    def __init__(
        self,
        soma: SomaConnect,
        mac: str,
        name: str,
        type: str,  # pylint: disable=redefined-builtin
        gen: str,
    ) -> None:
        """Initialise the Shade."""
        self._soma = soma
        self._mac = mac
        self._name = name
        self._type: str | None = type.capitalize() if type else None
        self._gen: str | None = gen or None
        self._position: int | None = None
        self._battery_level: int | None = None
        self._battery_percentage: int | None = None
        self._light_level: int | None = None
        self._light_level_last_updated: datetime.datetime | None = None

    def __str__(self) -> str:
        """Return a string representation of the shade."""
        return f"{self.name}: {self.model[0]} {self.model[1]} ({self.mac})"

    def __repr__(self) -> str:
        """Return representation of the shade."""
        return f"<{self.name} {self.model[0]} {self.model[1]} ({self.mac})>"

    def __hash__(self) -> int:
        """Return an integer hash for the mac address of the shade."""
        return hash(self._mac)

    @property
    def name(self) -> str:
        """Return the name of the shade."""
        return self._name

    @property
    def mac(self) -> str:
        """Return the mac address of the shade."""
        return self._mac

    @property
    def model(self) -> tuple[str | None, str | None]:
        """Return the model (type, gen) of the shade."""
        return (self._type, self._gen)

    @property
    def position(self) -> int | None:
        """Return the current position."""
        if self._position is not None:
            return int(self._position)
        return None

    @property
    def battery_level(self) -> int | None:
        """Return the current battery level."""
        if self._battery_level is not None:
            return int(self._battery_level)
        return None

    @property
    def battery_percentage(self) -> int | None:
        """Return the current battery level."""
        if self._battery_percentage is not None:
            return int(self._battery_percentage)
        return None

    @property
    def light_level(self) -> int | None:
        """Return the current light level."""
        if self._light_level is not None:
            return int(self._light_level)
        return None

    async def open(self) -> bool:
        """Open this shade."""
        return await self._soma.open_shade(self._mac)

    async def close(self) -> bool:
        """Close shade."""
        return await self._soma.close_shade(self._mac)

    async def stop(self) -> bool:
        """Stop shade."""
        return await self._soma.stop_shade(self._mac)

    async def set_position(
        self, position: int, close_upwards: bool = False, morning_mode: bool = False
    ) -> bool:
        """Set shade to specific position."""
        result = await self._soma.set_shade_position(
            self._mac,
            position,
            close_upwards=close_upwards,
            morning_mode=morning_mode,
        )
        if result is True:
            self._position = position

        return result

    async def get_current_position(self) -> int | None:
        """Update and return shade position."""
        response = await self._soma.get_shade_position(self._mac)
        if response is not False:
            self._position = response
            return self.position
        return None

    async def get_current_battery_level(self) -> int | None:
        """Get battery level."""
        response = await self._soma.get_battery_level(self._mac)
        if isinstance(response, tuple):
            self._battery_level = response[0]
            self._battery_percentage = response[1]
            return self.battery_level
        return None

    async def get_current_light_level(self) -> int | None:
        """
        Asks SOMA Connect to retrieve the light level from the motor at
        most once every ten minutes.

        The rate limiting is done to minimise the impact on the battery life
        of each motor as the request has to be made using an active Bluetooth
        connection.
        """

        # only update the light level value if there is no current value
        # or if there is a value, at most once every 10 minutes
        if (
            self._light_level is None
            or self._light_level_last_updated is None
            or (
                datetime.datetime.now() - self._light_level_last_updated
                > datetime.timedelta(minutes=10)
            )
        ):
            light_level = await self._soma.get_light_level(self._mac)
            self._light_level = light_level
            self._light_level_last_updated = datetime.datetime.now()

        return self.light_level



"""aiosoma CLI."""
from __future__ import annotations

import asyncio
import socket
from functools import wraps
from typing import Any, Iterable

import click
from rich import box
from rich import print as markup
from rich.console import Console
from rich.pretty import pprint
from rich.table import Table

from aiosoma import SomaConnect, SomaShade

JSON_OUTPUT: bool = False
LIVE_UPDATE: bool = True


def coro(func: Any) -> Any:
    """Function annotation to mark a command as an async coroutine."""

    @wraps(func)
    def wrapper(*args: Iterable[Any], **kwargs: dict[str, Any]) -> Any:
        return asyncio.run(func(*args, **kwargs))

    return wrapper


async def get_shade(
    soma: SomaConnect, mac: str | None = None, name: str | None = None
) -> SomaShade | None:
    """Get SomaShade from provided parameters."""

    if mac is None and name is None:
        raise click.UsageError("Must provide at least one of --mac or --name")

    if mac is not None and name is not None:
        raise click.UsageError("Must provide only one of --mac or --name")

    if mac is not None:
        return await get_shade_from_mac(soma, mac)

    if name is not None:
        return await get_shade_from_name(soma, name)

    return None


async def get_shade_from_mac(soma: SomaConnect, mac: str) -> SomaShade | None:
    """Return shade with specified mac"""
    response = await soma.list_devices()
    if isinstance(response, dict):
        for shade in response.get("shades", {}):
            if shade["mac"] == mac:
                return SomaShade(soma, **shade)
    return None


async def get_shade_from_name(soma: SomaConnect, name: str) -> SomaShade | None:
    """Return shade with specified name"""
    response = await soma.list_devices()
    if isinstance(response, dict):
        for device in response.get("shades", {}):
            if device["name"] == name:
                return SomaShade(soma, **device)
    return None


def handle_result(result: dict[str, Any] | None) -> None:
    """Handle the result."""
    if JSON_OUTPUT is True:
        pprint(result)
    elif result is None:
        markup("response: [bold red]error[/bold red]")
    elif isinstance(result, dict) and result.get("result", "") == "success":
        markup("response: [bold green]success[/bold green]")


@click.group()
@click.option(
    "--host",
    default="soma-connect",
    help="Host name or IP address of SOMA SomaConnect.",
    type=str,
)
@click.option("--port", default=3000, help="API port of SOMA SomaConnect", type=int)
@click.option(
    "--json", is_flag=True, default=False, help="Return result in JSON format"
)
@click.option(
    "--live", is_flag=True, default=True, help="Query the device for current value."
)
@click.pass_context
def cli(context: click.Context, host: str, port: int, json: bool, live: bool) -> None:
    """Simple soma CLI interface to control shades."""
    global JSON_OUTPUT, LIVE_UPDATE  # pylint: disable=global-statement

    context.obj = SomaConnect(host, port)
    JSON_OUTPUT = json
    LIVE_UPDATE = live

    try:
        ip_addr = socket.gethostbyname(host)
        _ = socket.create_connection((ip_addr, port), 1)

    except socket.gaierror as exc:
        raise click.UsageError(f"{host} not found") from exc


@cli.command()
@click.pass_obj
@coro
async def list_shades(soma: SomaConnect) -> None:
    """Output the list of shades discovered by SOMA SomaConnect."""

    response = await soma.list_devices()
    if isinstance(response, dict):
        if JSON_OUTPUT is True:
            pprint(response)
        else:
            console = Console()
            table = Table("name", "mac ", "model", "gen", box=box.ASCII)
            for shade in response["shades"]:
                _shade = SomaShade(soma, **shade)
                table.add_row(_shade.name, _shade.mac, _shade.model[0], _shade.model[1])
            console.print(table)


@cli.command()
@click.option("--mac", "-m", help="Mac address of the shade.", type=str)
@click.option("--name", "-n", help="Name of the shade.", type=str)
@click.pass_obj
@coro
async def open_shade(soma: SomaConnect, mac: str | None, name: str | None) -> None:
    """Open a shade by providing either its MAC address or name."""

    shade: SomaShade | None = await get_shade(soma, mac, name)
    if shade is not None:
        handle_result(await shade.open())


@cli.command()
@click.option("--mac", "-m", help="Mac address of the shade", type=str)
@click.option("--name", "-n", help="Name of the shade", type=str)
@click.pass_obj
@coro
async def close_shade(soma: SomaConnect, mac: str | None, name: str | None) -> None:
    """Close a shade by providing either its MAC address or name"""

    shade: SomaShade | None = await get_shade(soma, mac, name)
    if shade is not None:
        handle_result(await shade.close())


@cli.command()
@click.option(
    "--position",
    required=True,
    type=click.IntRange(0, 100, clamp=True),
    help="Target position (0-100)",
)
@click.option(
    "--upwards", is_flag=True, default=False, help="Close upwards (requires Tilt)"
)
@click.option(
    "--slow/--fast",
    default=False,
    help="Enable morning mode (quieter and slower)",
)
@click.option("--mac", "-m", help="Mac address of the shade", type=str)
@click.option("--name", "-n", help="Name of the shade", type=str)
@click.pass_obj
@coro
async def set_position(
    soma: SomaConnect,
    position: int,
    upwards: bool,
    slow: bool,
    mac: str | None,
    name: str | None,
) -> None:
    """
    Set a shade to specific position

    Includes support for opening Tilt shades upwards and for enabling morning
    mode which makes the shade quieter but slower.
    """
    shade: SomaShade | None = await get_shade(soma, mac, name)
    kwargs: dict[str, bool] = {"close_upwards": False, "morning_mode": False}
    if upwards is True:
        kwargs["close_upwards"] = True
    if slow is True:
        kwargs["morning_mode"] = True
    if shade is not None:
        handle_result(await shade.set_position(position, **kwargs))


@cli.command
@click.option("--mac", "-m", help="Mac address of the shade", type=str)
@click.option("--name", "-n", help="Name of the shade", type=str)
@click.pass_obj
@coro
async def stop_shade(soma: SomaConnect, mac: str | None, name: str | None) -> None:
    """
    Stop all motion on the shade
    """
    shade: SomaShade | None = await get_shade(soma, mac, name)
    if shade is not None:
        handle_result(await shade.stop())


@cli.command
@click.option("--mac", "-m", help="Mac address of the shade", type=str)
@click.option("--name", "-n", help="Name of the shade", type=str)
@click.pass_obj
@coro
async def get_position(soma: SomaConnect, mac: str | None, name: str | None) -> None:
    """Return the current state of a shade"""

    shade: SomaShade | None = await get_shade(soma, mac, name)
    if isinstance(shade, SomaShade):
        result = await shade.get_current_position()
        if isinstance(result, dict):
            if JSON_OUTPUT is True:
                pprint({"name": shade.name, "mac": shade.mac, "position": result})
            else:
                pprint(f"{shade.name} ({shade.mac}) position: {result}")


@cli.command
@click.option("--mac", "-m", help="Mac address of the shade.", type=str)
@click.option("--name", "-n", help="Name of the shade.", type=str)
@click.pass_obj
@coro
async def get_battery_level(
    soma: SomaConnect, mac: str | None, name: str | None
) -> None:
    """Return the current state of the shade."""

    shade: SomaShade | None = await get_shade(soma, mac, name)
    if isinstance(shade, SomaShade):
        result = await shade.get_current_battery_level()
        if isinstance(result, dict):
            if JSON_OUTPUT is True:
                pprint(
                    {
                        "name": shade.name,
                        "mac": shade.mac,
                        "battery_level": shade.battery_level,
                        "battery_percentage": shade.battery_percentage,
                    }
                )
            else:
                _battery_level = shade.battery_level or "n/a"
                _battery_percentage = shade.battery_percentage or "n/a"
                pprint(f"{shade.name} ({shade.mac}) level: {_battery_level}")
                pprint(f"{shade.name} ({shade.mac}) percentage: {_battery_percentage}")


@cli.command
@click.option("--mac", "-m", help="Mac address of the shade", type=str)
@click.option("--name", "-n", help="Name of the shade", type=str)
@click.pass_obj
@coro
async def get_light_level(soma: SomaConnect, mac: str | None, name: str | None) -> None:
    """Return the current state of a shade"""

    shade: SomaShade | None = await get_shade(soma, mac, name)
    if isinstance(shade, SomaShade):
        result = await shade.get_current_light_level()
        if isinstance(result, dict):
            if JSON_OUTPUT is True:
                pprint(result)
            else:
                _shade = SomaShade(
                    soma,
                    shade.mac,
                    shade.name,
                    shade.model,
                    shade.gen,
                    light_level=result["light_level"],
                )
                _light_level = (
                    f"{_shade.light_level}" if _shade.light_level is not None else "n/a"
                )
                console = Console()
                table = Table(show_header=False, box=box.ASCII)
                table.add_row("name", _shade.name)
                table.add_row("mac", _shade.mac)
                table.add_row("light level", _light_level)
                console.print(table)
